import subprocess
import sys
import importlib.util

def install_pip3():
    try:
        subprocess.check_call([sys.executable, "-m", "pip", "--version"])
        print("pip3 is already installed.")
    except subprocess.CalledProcessError:
        print("pip3 is not installed. Installing pip3...")
        subprocess.check_call([sys.executable, "-m", "ensurepip","--upgrade"])
        print("pip3 has been installed.")

def is_installed(package_name):
    package_spec = importlib.util.find_spec(package_name)
    return package_spec is not None

def install_packages(packages):
    for package in packages:
        if not is_installed(package):
            print(f"Installing {package}...")
            subprocess.check_call([sys.executable, "-m", "pip", "install", package])
        else:
            print(f"{package} is already installed.")

if __name__ == "__main__":
    install_pip3()
    packages = ["argparse","anndata","pandas","numpy","pydeseq2","scipy"]
    install_packages(packages)

import argparse, anndata, os,shutil,sys, warnings, subprocess
import pandas as pd
import numpy as np
from pydeseq2.dds import DeseqDataSet
from pydeseq2.ds import DeseqStats
from scipy.sparse import csr_matrix
pd.options.mode.chained_assignment = None
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    warnings.warn("deprecated", DeprecationWarning)


################# parser block
parser = argparse.ArgumentParser(description="Description of your program")
parser.add_argument("-f", "--file", type=str, default=None, help="dataset file name", metavar="character")
parser.add_argument("-c", "--ctrl", type=str, default=None, help="control dataset file name", metavar="character")
parser.add_argument("-o", "--org", type=str, default=None, help="dataset organism: human, mouse, yeast, fly", metavar="character")
parser.add_argument("-a", "--annot", type=str, default=None, help="name of the variable containing the annotation of cells", metavar="character")
parser.add_argument("-g", "--geno", type=str, default=None, help="name of the variable containing the genotype", metavar="character")
parser.add_argument("-n", "--name", type=str, default=None, help="file output name", metavar="character")
parser.add_argument("-t", "--thresh", type=int, default=30, help="low count threshold for pseudo bulk", metavar="integer")

args = parser.parse_args()
filename, extension = os.path.splitext(args.file)


if extension == '.rds':
     if args.file is None or args.org is None :
         parser.print_help()
         parser.error("At least 2 arguments must be supplied (input file) and(organism)")
else :
    if args.file is None or args.org is None or args.annot is None :
        parser.print_help()
        parser.error("At least 3 arguments must be supplied (input file), (organism), and (annot)")

    if args.org not in ['human', 'mouse', 'fly', 'yeast']:
        parser.print_help()
        parser.error("Only 4 organisms are available: human, mouse, fly, yeast")

if args.name is None:
    output_name = 'input_to_Xplorer'
    output_name_wt = 'WT_input_to_Xplorer'
    output_name_wt = 'MT_input_to_Xplorer'
    csv_name = 'Xplorer_pseudo_bulk.csv'
    csv_name_wt = 'WT_Xplorer_pseudo_bulk.csv'
    csv_name_mt = 'MT_Xplorer_pseudo_bulk.csv'
    csv_input_name = 'bulkXplorer.csv'
    csv_input_name = 'WT_bulkXplorer.csv'
    csv_input_name = 'MT_bulkXplorer.csv'
else:
    output_name = args.name
    output_name_wt = 'WT_' + args.name
    output_name_mt = 'MT_' + args.name
    csv_name = output_name + '_pseudo_bulk.csv'
    csv_name_wt = output_name_wt + '_pseudo_bulk.csv'
    csv_name_mt = output_name_mt + '_pseudo_bulk.csv'
    csv_input_name = output_name + '_bulkXplorer.csv'
    csv_input_name_wt = output_name_wt + '_bulkXplorer.csv'
    csv_input_name_mt = output_name_mt + '_bulkXplorer.csv'

###################
# adds vars Cell_Type, org and Cluster_ID that can be read by subUmap.py and in jsons in mx3
# and filter mito genes
def mito_filter(object,annot, org):
        ob = anndata.read_h5ad(object)
        ob.obs['Cell_Type'] = ob.obs[annot]
        ob.obs['org'] = [org]*len(ob.obs)
        ctdf = pd.DataFrame({'Cluster_ID': range(1, len(ob.obs['Cell_Type'].cat.categories) + 1)},index=ob.obs['Cell_Type'].cat.categories)
        ob.obs['Cluster_ID'] = ctdf.loc[ob.obs['Cell_Type'], 'Cluster_ID'].values
        ob.var['Countcol'] =  range(len(ob.var))
        if args.geno is None:
            ob.obs = pd.DataFrame({'Cell_Type' : ob.obs.Cell_Type.values, 'Cluster_ID' : ob.obs.Cluster_ID.values, 'org' : ob.obs.org.values}, index = ob.obs_names)
        else :
            ob.obs = pd.DataFrame({'Cell_Type' : ob.obs.Cell_Type.values, 'Cluster_ID' : ob.obs.Cluster_ID.values, 'org' : ob.obs.org.values, 'Cond' : ob.obs['disease']}, index = ob.obs_names)
        ob.var = pd.DataFrame({'Countcol' : ob.var.Countcol.values}, index = ob.var_names)
        del ob.obsp, ob.uns, ob.varm
        ad = ob[:,ob.var_names.isin(genes_to_filter_from[org])]
        return ob,ad

# creates a metadata df
def metadata_gen(anndata_object):
    umap_coord = pd.DataFrame(anndata_object.obsm['X_umap'], columns=['umap_1', 'umap_2'])
    metadata = pd.DataFrame(anndata_object.obs['Cell_Type'])
    metadata['Cluster_ID'] = anndata_object.obs['Cluster_ID']
    umap_coord.index = metadata.index
    metadata['UMAP1'] = umap_coord['umap_1']
    metadata['UMAP2'] = umap_coord['umap_2']
    metadata.reset_index(inplace = True)
    metadata.columns = ['Seq_ID', 'Cell_Type','Cluster_ID', 'UMAP1', 'UMAP2']
    col_order = ['Cell_Type','Seq_ID','Cluster_ID', 'UMAP1', 'UMAP2']
    metadata = metadata[col_order]
    del umap_coord
    return metadata

# creates an object per cell type and put them into a list
def get_unique_cluster(pbmetadata, pbtr_counts) :
    unique_clusters = sorted(pbmetadata['Cluster_ID'].unique())
    for cluster in unique_clusters:
        cluster_df = pbtr_counts[pbtr_counts['Cluster_ID'] == cluster]
        cluster_df['Seq_ID'] = pbmetadata['Seq_ID']
        cluster_df.set_index('Seq_ID', inplace = True)
        nam = f"cluster0{cluster}" if cluster < 10 else f"cluster{cluster}"
        globals()[nam] = cluster_df.copy()
    del cluster_df,cluster
    variable_names = [var for var in globals() if var.startswith('cluster')]
    dflst = [globals()[var] for var in variable_names]
    dflst = [df.iloc[:, 1:] for df in dflst]
    return dflst

# generation of the pseudo bulk from raw counts
def pbgeneration(ob) :
    pbmetadata = metadata_gen(ob)
    pbdata = ob.raw.X############
    pbdata = pd.DataFrame(pbdata.toarray())
    pbdata.columns = ob.raw.var_names ####
    pbtr_counts = pd.concat([pbmetadata['Cluster_ID'], pbdata], axis=1)
    pbtr_counts.index = ob.obs_names
    pbtr_counts.reset_index(inplace = True)
    pbtr_counts = pbtr_counts.rename(columns = {'cell_id':'Seq_ID'})
    dflst = get_unique_cluster(pbmetadata,pbtr_counts)
    cpt = 0
    genes = pbdata.columns
    final_pseudo_bulk = pd.DataFrame(index = genes)
    pseudo_bulk_tot = pd.DataFrame(index = genes)
    for celltype in range(len(dflst)):
        cl = dflst[celltype].transpose()
        if cl.shape[1] == 1:
            sums = np.round(cl.sum(axis=1))
            cl.insert(1, 'B', cl.iloc[:, 0])
            cl.insert(2, 'C', cl.iloc[:, 0])
            cl.insert(3, 'D', cl.iloc[:, 0])
        elif cl.shape[1] == 2:
            sums = np.round(cl.sum(axis=1))
            cl.insert(2, 'C', cl.iloc[:, 0])
            cl.insert(3, 'D', cl.iloc[:, 1])
        elif cl.shape[1] == 3:
            sums = np.round(cl.sum(axis=1))
            cl.insert(3, 'D', cl.iloc[:, 0])
        else:
            sums = np.round(cl.sum(axis=1))
        mid1 = cl.shape[1] // 3
        mid2 = 2 * cl.shape[1] // 3
        one = np.round(cl.iloc[:, :mid1].sum(axis=1))
        two = np.round(cl.iloc[:, mid1:mid2].sum(axis=1))
        three = np.round(cl.iloc[:, mid2:].sum(axis=1))
        name_1 = f"{pbmetadata['Cell_Type'].cat.categories[cpt]}_1"
        name_2 = f"{pbmetadata['Cell_Type'].cat.categories[cpt]}_2"
        name_3 = f"{pbmetadata['Cell_Type'].cat.categories[cpt]}_3"
        name_1 = name_1.replace(', ', '_').replace(' ', '.')
        name_2 = name_2.replace(', ', '_').replace(' ', '.')
        name_3 = name_3.replace(', ', '_').replace(' ', '.')
        name = pbmetadata['Cell_Type'].cat.categories[cpt]
        name = name.replace(', ', '_').replace(' ', '.')
        pseudo_bulk = pd.DataFrame({name_1: one, name_2: two, name_3: three}, index=genes)
        pb_2 = pd.DataFrame({name: sums}, index=genes)
        final_pseudo_bulk = pd.concat([final_pseudo_bulk, pseudo_bulk], axis=1)
        pseudo_bulk_tot = pd.concat([pseudo_bulk_tot, pb_2], axis=1)
        cpt += 1
    return pseudo_bulk_tot, final_pseudo_bulk

# manages the pseudo bulk of the wt vs mt
def pb_wtmt(pseudo_bulk_tot, final_pseudo_bulk) :
    final_pseudo_bulk = final_pseudo_bulk*3
    final_pseudo_bulk.reset_index(inplace = True)
    final_pseudo_bulk = final_pseudo_bulk.rename(columns={final_pseudo_bulk.columns[0]: 'genes'})
    pseudo_bulk_tot = pseudo_bulk_tot.rename(columns={pseudo_bulk_tot.columns[0]: 'genes'})
    return pseudo_bulk_tot, final_pseudo_bulk

# manages to pseudo bulk of a cell type against the mean
def pb_one_cond(pseudo_bulk_tot, final_pseudo_bulk):
    final_pseudo_bulk = final_pseudo_bulk*3 #to get realistic values regarding the real values and also as we need to round the means fro deseq2, this is the only way to not miss some genes
    final_pseudo_bulk['mean_rep1'] = np.round(final_pseudo_bulk.filter(like='_1').sum(axis=1) / ((final_pseudo_bulk.shape[1]) / 2))
    final_pseudo_bulk['mean_rep2'] = np.round(final_pseudo_bulk.filter(like='_2').sum(axis=1) / ((final_pseudo_bulk.shape[1]-1) / 2))
    final_pseudo_bulk['mean_rep3'] = np.round(final_pseudo_bulk.filter(like='_3').sum(axis=1) / ((final_pseudo_bulk.shape[1]-2) / 2))
    final_pseudo_bulk.reset_index(inplace = True)
    final_pseudo_bulk = final_pseudo_bulk.rename(columns={final_pseudo_bulk.columns[0]: 'genes'})
    pseudo_bulk_tot['mean'] = pseudo_bulk_tot.sum(axis=1)/(pseudo_bulk_tot.shape[1])
    pseudo_bulk_tot.reset_index(inplace = True)
    pseudo_bulk_tot = pseudo_bulk_tot.rename(columns={pseudo_bulk_tot.columns[0]: 'genes'})
    return pseudo_bulk_tot, final_pseudo_bulk



# performs pydeseq2 when 2 conditions were given
def bulkXplorer_two_files(cond_final,cond_pb, wt_final,wt_pb,thresh) :
    pb_input = pd.DataFrame(columns = ['CellType', 'genes','control','mutant','log2fc','pval'])
    df_wt = []
    df_mt = []
    for i in range(1, len(wt_final.columns), 3):
        if i + 2 < len(wt_final.columns):
            df_temp = wt_final.iloc[:, [i, i+1, i+2]]
            df_wt.append(df_temp)
    for i in range(1, len(cond_final.columns), 3):
        if i + 2 < len(cond_final.columns):
            df_temp2 = cond_final.iloc[:, [i, i+1, i+2]]
            df_mt.append(df_temp2)
        #here for each cell type we compute the log2fc and the pvalue comparing each cell type to the mean expression across all cells
    for ct in range(wt_pb.shape[1]-2):
        df = pd.DataFrame({'CellType' : wt_pb.columns[ct+1], 'genes': wt_pb['genes'], 'mutant' : cond_pb.iloc[:,ct+1], 'control': wt_pb.iloc[:,ct+1]})
        counts = pd.DataFrame({'genes': wt_pb['genes'], 'M_1' :df_mt[ct].iloc[:,0].values,'M_2': df_mt[ct].iloc[:,1].values,'M_3': df_mt[ct].iloc[:,2].values,'C_1': df_wt[ct].iloc[:,0].values,'C_2' : df_wt[ct].iloc[:,1].values,'C_3' : df_wt[ct].iloc[:,2].values})
        counts = counts.set_index('genes')
        counts = counts[counts.sum(axis=1) >= thresh]
        df = df[df['genes'].isin(counts.index)]
        df = df.reset_index()
        df = df.drop(df.columns[0], axis=1)
        counts = counts.T
        metadata = pd.DataFrame(zip(counts.index,['M','M','M','C','C','C']),columns=['Sample','Condition'])
        metadata = metadata.set_index('Sample')
        dds = DeseqDataSet(counts = counts, metadata = metadata, design_factors = 'Condition', quiet = True)
        dds.deseq2()
        stat_res = DeseqStats(dds, contrast = ('Condition','M','C'),quiet = True)
        stat_res.summary(quiet = True)
        res = stat_res.results_df
        log2 = res['log2FoldChange']
        log2.reset_index(drop = True, inplace = True)
        pval = res['pvalue']
        pval.reset_index(drop = True, inplace = True)
        df['log2fc'] = log2
        df['pval'] = pval
        nor_counts = pd.DataFrame(dds.layers['normed_counts'])
        nor_counts = nor_counts.T
        mt = nor_counts[[0, 1, 2]].sum(axis=1)
        wt = nor_counts[[3, 4, 5]].sum(axis=1)
        df['control'] = wt
        df['mutant'] = mt
        pb_input = pd.concat([pb_input, df], ignore_index=True)
    pb_input = pb_input.dropna(subset = ['log2fc'])
    return pb_input



# performs pydeseq2 when against the mean
def bulkXplorer_one_file(final,pb,thresh):
    pb_input = pd.DataFrame(columns = ['CellType', 'genes','control','mutant','log2fc','pval'])
    dfs = []
    for i in range(1, len(final.columns), 3):
        if i + 2 < len(final.columns):
            df_temp = final.iloc[:, [i, i+1, i+2]]
            dfs.append(df_temp)
    dfs.pop()
    #here for each cell type we compute the log2fc and the pvalue comparing each cell type to the mean expression across all cells
    for ct in range(pb.shape[1]-2):
        df = pd.DataFrame({'CellType' : pb.columns[ct+1], 'genes': pb['genes'], 'mutant' : pb.iloc[:,ct+1], 'control': pb.iloc[:,ct+1]})
        counts = pd.DataFrame({'genes': pb['genes'], 'M_1' :dfs[ct].iloc[:,0].values,'M_2': dfs[ct].iloc[:,1].values,'M_3': dfs[ct].iloc[:,2].values,'C_1': final['mean_rep1'],'C_2' : final['mean_rep2'],'C_3' : final['mean_rep3']})
        counts = counts.set_index('genes')
        counts = counts[counts.sum(axis=1) >= thresh]
        df = df[df['genes'].isin(counts.index)]
        df = df.reset_index()
        df = df.drop(df.columns[0], axis=1)
        counts = counts.T
        metadata = pd.DataFrame(zip(counts.index,['M','M','M','C','C','C']),columns=['Sample','Condition'])
        metadata = metadata.set_index('Sample')
        dds = DeseqDataSet(counts = counts, metadata = metadata, design_factors = 'Condition',quiet = True)
        dds.deseq2()
        stat_res = DeseqStats(dds, contrast = ('Condition','M','C'), quiet = True)
        stat_res.summary(quiet = True)
        res = stat_res.results_df
        log2 = res['log2FoldChange']
        log2.reset_index(drop = True, inplace = True)
        pval = res['pvalue']
        pval.reset_index(drop = True, inplace = True)
        df['log2fc'] = log2
        df['pval'] = pval
        nor_counts = pd.DataFrame(dds.layers['normed_counts'])
        nor_counts = nor_counts.T
        mt = nor_counts[[0, 1, 2]].sum(axis=1)
        wt = nor_counts[[3, 4, 5]].sum(axis=1)
        df['control'] = wt
        df['mutant'] = mt
        pb_input = pd.concat([pb_input, df], ignore_index=True)
    pb_input = pb_input.dropna(subset = ['log2fc'])
    return pb_input

# saves mx3 object
def save_mx3(ad, op_name):
    final_output = op_name+'.h5ad'
    ad.write_h5ad(final_output)
    shutil.make_archive(op_name, 'zip', scXplorer_directory, final_output)
    os.remove(final_output)

#saves pseudo bulk
def save_mx2(final, pb_input, csvName,inputName):
    final.to_csv(csvName, index=False)
    pb_input.to_csv(inputName, index=False)

#performs everything when 2 conditions were given but only one object
def ending(thresh):
    save_mx3(wt,output_name_wt)
    save_mx3(mt, output_name_mt)
    save_mx3(ad, output_name)
    print('Inputs to MitoXplorer3 are ready, pseudo bulks are in generation...')
    pb_wt, final_wt = pbgeneration(ob_wt)
    pb_mt, final_mt = pbgeneration(ob_mt)
    pb_mt1, final_mt1 = pb_one_cond(pb_mt, final_mt)
    pb_wt1, final_wt1 = pb_one_cond(pb_wt,final_wt)
    pbwt, finalwt = pb_wtmt(pb_wt,final_wt)
    pbmt, finalmt = pb_wtmt(pb_mt, final_mt)
    pb_input = bulkXplorer_two_files(finalmt,pbmt,finalwt,pbwt,thresh)
    pb_input_wt = bulkXplorer_one_file(final_wt1, pb_wt1,thresh)
    save_mx2(final_wt1,pb_input_wt,csv_name_wt, csv_input_name_wt)
    pb_input_mt = bulkXplorer_one_file(final_mt1,pb_mt1, thresh)
    save_mx2(final_mt1,pb_input_mt,csv_name_mt, csv_input_name_mt)

    pb_input.to_csv(csv_input_name, index=False)

# uses the R script to format the rds data for anndata and scanpy and creates the object
def rdsProcess(file):
    Rcommand = ["Rscript", "../subproR.R", '../'+file]
    subprocess.run(Rcommand, check=True)
    metadata = pd.read_csv('metadata.csv',header =0, index_col = 0)
    counts = pd.read_csv('counts.csv',header = 0, index_col = 0)
    var = counts.columns
    matrix = csr_matrix(counts.astype(pd.SparseDtype("float64",0)).sparse.to_coo())
    adata = anndata.AnnData(X=matrix, obs = metadata)
    adata.var_names = var
    adata.obsm['X_umap'] = metadata[['UMAP1', 'UMAP2']].to_numpy(dtype=np.float32)
    adata.obs['Cell_Type'] = adata.obs['Cell_Type'].astype('category')
    os.remove('counts.csv')
    os.remove('metadata.csv')
    return adata

####code :
genes_to_filter_from = pd.read_csv('./mitogenes.csv',sep='\t')

current_working_directory = os.getcwd()
scXplorer_directory = os.path.join(current_working_directory, 'scXplorer')
if not os.path.exists(scXplorer_directory):
    os.makedirs(scXplorer_directory)
os.chdir(scXplorer_directory)


if extension == '.rds' :
    if args.ctrl is None :
        ob = rdsProcess(args.file)
        ad = ob[:,ob.var_names.isin(genes_to_filter_from[args.org])]
        save_mx3(ad,output_name)
        print('Input to MitoXplorer3 is ready, pseudo bulk is in generation...')
        pb,final = pbgeneration(ob)
        pb,final = pb_one_cond(pb,final)
        pb_input = bulkXplorer_one_file(final,pb, int(args.thresh))
        save_mx2(final,pb_input,csv_name, csv_input_name)
    else:
        filename_wt, extension_wt = os.path.splitext(args.ctrl)
        if extension_wt == '.rds' : #if a mt AND wt file
            ob_mt = rdsProcess(args.file)
            mt = ob_mt[:,ob_mt.var_names.isin(genes_to_filter_from[args.org])]
            ob_wt = rdsProcess(args.ctrl)
            wt = ob_wt[:,ob_wt.var_names.isin(genes_to_filter_from[args.org])]
            ad = anndata.concat([mt, wt])
            ending(int(args.thresh))

else :

    if args.ctrl is None:
        ob,ad = mito_filter('../'+args.file,args.annot,args.org)
        if args.geno is None:
            save_mx3(ad,output_name)
            print('Input to MitoXplorer3 is ready, pseudo bulk is in generation...')
            pb,final = pbgeneration(ob)
            pb,final = pb_one_cond(pb,final)
            pb_input = bulkXplorer_one_file(final,pb, int(args.thresh))
            save_mx2(final,pb_input,csv_name, csv_input_name)
        else:
            wts = ['WT','Wt','wt','Ctrl','ctrl','ctl','control','CTRL','normal','Normal','healthy','Healthy'] ##
            wt =  ad[ad.obs.Cond.isin(wts)]
            mt =  ad[~ad.obs.Cond.isin(wts)]
            ob_wt =  ob[ob.obs.Cond.isin(wts)]
            ob_mt =  ob[~ob.obs.Cond.isin(wts)]
            ending(int(args.thresh))
    else:
        ob_mt,mt = mito_filter('../'+args.file,args.annot,args.org)
        mt.obs['Cond'] = ['MT']*len(mt.obs)
        ob_wt,wt = mito_filter('../'+args.ctrl,args.annot,args.org)
        wt.obs['Cond'] = ['WT']*len(wt.obs)
        ad = anndata.concat([mt, wt])
        ending(int(args.thresh))
